CUST_TOASTS = "Toasts"
# CUST_TOASTS_ERROR = "ToastsError" --> Replaced by GC.TESTCASEERRORLOG
VIGOGFNUMMER = "GF#"
SAPPOLNR = "SAP Polizzennr"
PRAEMIE = "Prämie"
POLNRHOST = "PolNR Host"
VERMITTLER = "Vermittler"
MANDANT = "Mandant"
DONAU = "DON"
WSTV = "WSTV"
VN = "VN"

CLASSES_TESTCASE = "baangtVIG.CustTestCaseMaster.CustTestCaseMaster"